#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations


from .ElfObjDisasmInternals import getToolDescription as getToolDescription
from .ElfObjDisasmInternals import addOptionsToParser as addOptionsToParser
from .ElfObjDisasmInternals import getArgsParser as getArgsParser
from .ElfObjDisasmInternals import applyArgs as applyArgs
from .ElfObjDisasmInternals import applyGlobalConfigurations as applyGlobalConfigurations
from .ElfObjDisasmInternals import getOutputPath as getOutputPath
from .ElfObjDisasmInternals import getProcessedSections as getProcessedSections
from .ElfObjDisasmInternals import changeGlobalSegmentRanges as changeGlobalSegmentRanges
from .ElfObjDisasmInternals import insertSymtabIntoContext as insertSymtabIntoContext
from .ElfObjDisasmInternals import insertDynsymIntoContext as insertDynsymIntoContext
from .ElfObjDisasmInternals import injectAllElfSymbols as injectAllElfSymbols
from .ElfObjDisasmInternals import processGlobalOffsetTable as processGlobalOffsetTable
from .ElfObjDisasmInternals import processArguments as processArguments
from .ElfObjDisasmInternals import addSubparser as addSubparser
from .ElfObjDisasmInternals import elfObjDisasmMain as elfObjDisasmMain
