#!/usr/bin/env python3

# SPDX-FileCopyrightText: © 2022 Decompollaborate
# SPDX-License-Identifier: MIT

from __future__ import annotations


from .DisasmdisInternals import getToolDescription as getToolDescription
from .DisasmdisInternals import addOptionsToParser as addOptionsToParser
from .DisasmdisInternals import getArgsParser as getArgsParser
from .DisasmdisInternals import applyArgs as applyArgs
from .DisasmdisInternals import wordGeneratorFromStrList as wordGeneratorFromStrList
from .DisasmdisInternals import processArguments as processArguments
from .DisasmdisInternals import addSubparser as addSubparser
from .DisasmdisInternals import disasmdisMain as disasmdisMain
